/* drivers/barcelona/acc/acc.c
 *
 * Implementation of the accelerometer driver.
 *
 * Copyright (C) 2004,2005 TomTom BV <http://www.tomtom.com/>
 * Author: Dimitry Andric <dimitry.andric@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/* Includes */ 
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/wait.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/irq.h>
#include <asm/hardware.h>

/* Defines */
#define PFX "acc: "


#define ADC_RATE		100
#define ADC_CHANNELS	9
#define ADC_BUFSIZE		(ADC_CHANNELS * sizeof(short))
#define ADC_GYRO       4

#define ACC_DEVNAME					"acc"
#define ACC_MAJOR					120

typedef struct {
	unsigned int u32Gyro;					/* gyro data */
	unsigned int u32xData;				/* x axis data */
	unsigned int u32yData;				/* y axis data */
	unsigned int u32zData;				/* z axis data */
	unsigned int u32sTimeStamp;			/* timestamp */
	unsigned int u32usTimeStamp;			/* microseconds */
	unsigned int u32Temperature;			/* temperature */
} ACCMETER_DATA;

#define ACC_DRIVER_MAGIC			'A'
#define IOR_FIFOFILLED_SIZE			_IOR(ACC_DRIVER_MAGIC, 0, unsigned)		/* UNSUPPORTED */
#define IOW_ACC_SAMPLINGRATE		_IOW(ACC_DRIVER_MAGIC, 1, unsigned)

/* For IOW_ACC_SAMPLINGRATE ioctl */
#define ACC_RATE_MIN				1
#define ACC_RATE_DEF				25
#define ACC_RATE_MAX				100


#define ACC_FIFO_SIZE 256

#define ACC_SCALER_MIN (ADC_RATE / ACC_RATE_MAX)
#define ACC_SCALER_DEF (ADC_RATE / ACC_RATE_DEF)
#define ACC_SCALER_MAX (ADC_RATE / ACC_RATE_MIN)

/* Local variables */
static wait_queue_head_t acc_wait;
static volatile int acc_read_index;
static volatile int acc_write_index;
static ACCMETER_DATA acc_fifo[ACC_FIFO_SIZE];
static volatile unsigned acc_scaler = ACC_SCALER_DEF;

static void acc_tasklet(unsigned long data);
DECLARE_TASKLET(acc_tasklet_handle, acc_tasklet, 0L);

static unsigned short buf[ADC_CHANNELS];

static void (*acc_get_values)(short buf[ADC_CHANNELS],ACCMETER_DATA *acc_data) = 0;

#if 0
static void acc_adc_poll(short passedbuf[ADC_CHANNELS])
{
	memcpy(buf,passedbuf,sizeof(buf));
	tasklet_schedule(&acc_tasklet_handle);
}
#endif

static void acc_tasklet(unsigned long data)
{
	static unsigned count = 0;
	struct timeval tv;

	if (count > 0) {
		--count;
		return;
	}
	count = acc_scaler - 1;

	do_gettimeofday(&tv);

	/* Store accelerometer data in fifo. */

	/* Get acc values acoording to used type */
	if (acc_get_values) acc_get_values(buf,&acc_fifo[acc_write_index]);
	/* Get gyro data */
	acc_fifo[acc_write_index].u32Gyro        = buf[ADC_GYRO];
	/* Get timestamp */
	acc_fifo[acc_write_index].u32sTimeStamp  = tv.tv_sec;
	acc_fifo[acc_write_index].u32usTimeStamp = tv.tv_usec;

	if (++acc_write_index >= ACC_FIFO_SIZE)
		acc_write_index = 0;
	/* If fifo is full, bump read index. */
	if (acc_write_index == acc_read_index)
		if (++acc_read_index >= ACC_FIFO_SIZE)
			acc_read_index = 0;

	wake_up_interruptible(&acc_wait);
}

static ssize_t acc_read(struct file *file, char __user *data, size_t len, loff_t *ppos)
{
	ssize_t ret;

	/* Check user buffer sanity */
	if (len < sizeof(ACCMETER_DATA) || !access_ok(VERIFY_WRITE, data, sizeof(ACCMETER_DATA)))
		return -EFAULT;

	/* Check if data is available in the fifo */
	tasklet_disable(&acc_tasklet_handle);
	if (acc_write_index == acc_read_index) {
		tasklet_enable(&acc_tasklet_handle);
		if (file->f_flags & O_NONBLOCK) {
			return -EAGAIN;
		} else {
			DECLARE_WAITQUEUE(wait, current);

			add_wait_queue(&acc_wait, &wait);
			set_current_state(TASK_INTERRUPTIBLE);
			for (;;) {
				tasklet_disable(&acc_tasklet_handle);
				if (acc_write_index != acc_read_index)
					break;
				tasklet_enable(&acc_tasklet_handle);
				if (signal_pending(current)) {
					set_current_state(TASK_RUNNING);
					remove_wait_queue(&acc_wait, &wait);
					return -ERESTARTSYS;
				}
				schedule();
			}
			set_current_state(TASK_RUNNING);
			remove_wait_queue(&acc_wait, &wait);
		}
	}
	/* Spinlock should still be locked here... */
	ret = __copy_to_user(data, &acc_fifo[acc_read_index], sizeof(ACCMETER_DATA)) ? -EFAULT : sizeof(ACCMETER_DATA);
	if (++acc_read_index >= ACC_FIFO_SIZE)
		acc_read_index = 0;
	tasklet_enable(&acc_tasklet_handle);

	return ret;
}

static unsigned int acc_poll(struct file *file, struct poll_table_struct *wait)
{
	unsigned int ret;

	poll_wait(file, &acc_wait, wait);

	tasklet_disable(&acc_tasklet_handle);
	ret = (acc_write_index == acc_read_index) ? 0 : (POLLIN | POLLRDNORM);
	tasklet_enable(&acc_tasklet_handle);

	return ret;
}

static int acc_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int ret;
	unsigned rate;
	unsigned scaler;

	switch (cmd) {
	case IOR_FIFOFILLED_SIZE:
		ret = -EINVAL;
		break;
	case IOW_ACC_SAMPLINGRATE:
		if (copy_from_user(&rate, (void *) arg, sizeof(unsigned))) {
			ret = -EFAULT;
		} else if (rate < ACC_RATE_MIN || rate > ACC_RATE_MAX) {
			ret = -EINVAL;
		} else {
			scaler = ADC_RATE / rate;
			if (scaler < ACC_SCALER_MIN || scaler > ACC_SCALER_MAX) {
				ret = -EINVAL;
			} else {
				acc_scaler = scaler;
				ret = 0;
			}
		}
		break;
	default:

		ret = -EINVAL;
		break;
	}

	return ret;
}

static int acc_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

static int acc_release(struct inode *inode, struct file *file)
{
	return 0;
}

/* Kernel interface */
static struct file_operations acc_fops = {
	.owner		= THIS_MODULE,
	.read		= acc_read,
	.poll		= acc_poll,
	.ioctl		= acc_ioctl,
	.open		= acc_open,
	.release	= acc_release,
};

static void acc_hw_init(void)
{
}

static void acc_hw_exit(void)
{
}

static int acc_probe(void)
{
	int ret;

	init_waitqueue_head(&acc_wait);

	acc_hw_init();

	acc_get_values = 0;

#if 0
	ret = adc_register_poll(acc_adc_poll);
	if (ret < 0) {

		return ret;
	}
#endif

	ret = register_chrdev(ACC_MAJOR, ACC_DEVNAME, &acc_fops);
	if (ret < 0) {

		return ret;
	}


	return 0;
}

static int acc_remove(void)
{

	unregister_chrdev(ACC_MAJOR, ACC_DEVNAME);

#if 0
	adc_unregister_poll(acc_adc_poll);
#endif
	acc_hw_exit();


	return 0;
}

static int __init acc_mod_init(void)
{
  return acc_probe();
}

static void __exit acc_mod_exit(void)
{
  acc_remove();
}

module_init(acc_mod_init);
module_exit(acc_mod_exit);

MODULE_AUTHOR("Dimitry Andric <dimitry.andric@tomtom.com>");
MODULE_DESCRIPTION("TomTom GO Accelerometer Driver");
MODULE_LICENSE("GPL");

/* EOF */
